# awl.admintools.py
from django.urls import reverse
from django.template import Context, Template

# ============================================================================

def admin_obj_link(obj, display=''):
    """Returns a link to the django admin change list with a filter set to
    only the object given.

    :param obj:
        Object to create the admin change list display link for
    :param display:
        Text to display in the link.  Defaults to string call of the object
    :returns:
        Text containing HTML for a link
    """
    # get the url for the change list for this object
    url = reverse('admin:%s_%s_changelist' % (obj._meta.app_label,
        obj._meta.model_name))
    url += '?id__exact=%s' % obj.id

    text = str(obj)
    if display:
        text = display

    return '<a href="%s">%s</a>' % (url, text)


def make_admin_obj_mixin(name):
    """This method dynamically creates mixin to be used with your 
    :class:`ModelAdmin` classes that provides methods to display links to
    related objects.

    :param name:
        Each usage of the mixin must be given a unique name for the mixin class
        being created
    :returns:
        Dynamically created mixin class

    The created class will have a single method:

    .. code-block:: python

        add_obj_link(funcname, attr, [title, display])

    :param funcname:
        Name of the function to be added to the mixin.  This would normally
        correspond to an item in the :class:`ModelAdmin` class's
        ``list_display`` tuple
    :param attr:
        Name of the attribute belonging to the object instance of the model
        that the :class:`ModelAdmin` repsresents.  This essentially identifies
        what will be linked to
    :param title:
        Title for the column of the django admin table.  If not given it
        defaults to a capitalized version of ``attr``
    :param display:
        What to display as the text for the link being shown.  If not given it
        defaults to the string representation of the object for the row: 
        ``str(obj)`` .  This parameter supports django templating, the context
        for which contains a dictionary key named "obj" with the value being
        the object for the row.

    Example usage:

    .. code-block:: python

        # models file

        class Inner(models.Model):
            name = models.CharField(max_length=10)


        class Outer(models.Model):
            name = models.CharField(max_length=10)
            inner = models.ForeignKey(Inner, on_delete=models.CASCADE)


        # admin file
        @admin.register(Inner)
        class InnerAdmin(admin.ModelAdmin):
            list_display = ('name', )



        base = make_admin_obj_mixin('OuterMixin')
        base.add_obj_link('show_inner', 'inner', 'My Inner',
            'Inner.id={{obj.id}}')

        @admin.register(Outer)
        class OuterAdmin(admin.ModelAdmin):
            list_display = ('name', 'show_inner')


    A sample django admin page for "Outer" would have the table:

    +---------+-------------+
    | Name    | My Inner    |
    +=========+=============+
    | One     | Inner.id=1  |
    +---------+-------------+
    | Two     | Inner.id=2  |
    +---------+-------------+
    | Three   | Inner.id=3  |
    +---------+-------------+

    Each of the items in the "My Inner" column would be a link to the django
    admin change list for the "Inner" object with a filter set to show just
    the object that was clicked.
    """
    @classmethod
    def add_obj_link(cls, funcname, attr, title='', display=''):
        if not title:
            title = attr.capitalize()

        # python scoping is a bit weird with default values, if it isn't
        # reference the inner function won't see it, so assign it for use
        _display = display

        def _link(self, obj):
            # handle '__' referencing like in QuerySets
            fields = attr.split('__')
            field_obj = getattr(obj, fields[0])
            if not field_obj:
                return ''

            for field in fields[1:]:
                # keep going down the reference tree
                field_obj = getattr(field_obj, field, None)
                if not field_obj:
                    return ''

            link_name = ''
            if not _display:
                link_name = str(field_obj)
            else:
                template = Template(_display)
                context = Context({'obj':field_obj})
                link_name = template.render(context)

            return admin_obj_link(field_obj, link_name)
        _link.short_description = title
        _link.allow_tags = True
        _link.admin_order_field = attr

        setattr(cls, funcname, _link)

    klass = type(name, (), {})
    klass.add_obj_link = add_obj_link
    return klass



from django.db import models
from django.utils.encoding import python_2_unicode_compatible

from awl.absmodels import ValidatingMixin
from awl.rankedmodel.models import RankedModel

# ============================================================================
# Waelsteng Models
# ============================================================================

class Link(models.Model):
    url = models.CharField(max_length=80)
    text = models.CharField(max_length=80)


class Validator(ValidatingMixin, models.Model):
    counter = models.IntegerField(default=0)

    def full_clean(self):
        super(Validator, self).full_clean()
        self.counter += 1

# ============================================================================
# Admintools Models 
# ============================================================================

@python_2_unicode_compatible
class Nested(models.Model):
    name = models.CharField(max_length=10)

    def __str__(self):
        return 'Nested(id=%s %s)' % (self.id, self.name)


@python_2_unicode_compatible
class Inner(models.Model):
    name = models.CharField(max_length=10)
    nested = models.ForeignKey(Nested, null=True, blank=True,
        on_delete=models.CASCADE)

    def __str__(self):
        return 'Inner(id=%s %s)' % (self.id, self.name)

    @property
    def classname(self):
        return 'Inner'


class Outer(models.Model):
    name = models.CharField(max_length=10)
    inner = models.ForeignKey(Inner, null=True, blank=True,
        on_delete=models.CASCADE)

# ============================================================================
# RankedModel Models 

class Alone(RankedModel):
    name = models.CharField(max_length=1)

    def __init__(self, *args, **kwargs):
        # ignore fake group entry to make the constructors the same for both
        # test models
        kwargs.pop('group', None)
        super(Alone, self).__init__(*args, **kwargs)


class Grouped(RankedModel):
    group = models.CharField(max_length=1)
    name = models.CharField(max_length=1)

    def grouped_filter(self):
        return Grouped.objects.filter(group=self.group)



from django.contrib import messages
from django.http import HttpResponse
from django.template import Template, RequestContext

def test_view_for_messages(request):
    messages.success(request, 'One')
    messages.error(request, 'Two')

    t = Template('')
    context = RequestContext(request)
    response = HttpResponse(t.render(context))
    return response



import os, tempfile, shutil, mock, six, unittest
from django.contrib import messages
from django.test import TestCase, override_settings

from awl.waelsteng import AdminToolsMixin, messages_from_response

from awl.tests.admin import LinkAdmin
from awl.tests.models import Link
from awl.waelsteng import WRunner

# ============================================================================

class AdminToolsMixinTest(TestCase, AdminToolsMixin):
    def setUp(self):
        self.initiate()

    def test_mixin(self):
        link1 = Link.objects.create(url='/admin/', text='Admin')
        link2 = Link.objects.create(url='', text='Blank')
        link_admin = LinkAdmin(Link, self.site)

        self.authed_get('/admin/')
        self.authed_post('/admin/', {})

        expected = ('url', 'text', 'visit_me')
        self.assertEqual(expected, self.field_names(link_admin))
        self.assertEqual('Admin', self.field_value(link_admin, link1, 'text'))

        self.visit_admin_link(link_admin, link1, 'visit_me')
        with self.assertRaises(AttributeError):
            self.visit_admin_link(link_admin, link2, 'visit_me')


    def test_coverage(self):
        # miscellaneous pieces to get our coverage to 100%

        # test before get, to check that auth call worked
        self.authed_post('/admin/', {})

    def assert_messages(self, contents):
        self.assertEqual(2, len(contents))
        self.assertEqual('One', contents[0][0])
        self.assertEqual(messages.SUCCESS, contents[0][1])
        self.assertEqual('Two', contents[1][0])
        self.assertEqual(messages.ERROR, contents[1][1])

    def test_messages_from_response(self):
        response = self.authed_get('/awl_test_views/test_view_for_messages/')

        m = messages_from_response(response)
        self.assert_messages(m)

        # -- test handling with cookies

        # wipe out the context, so that the cookie that was set is what gets
        # processed
        response.context = None

        m = messages_from_response(response)
        self.assert_messages(m)

        # -- test handling bad response objects
        response.cookies.pop('messages')
        m = messages_from_response(response)
        self.assertEqual(0, len(m))

        # object without context
        m = messages_from_response({})
        self.assertEqual(0, len(m))

        # object with context but empty
        class Dummy(object):
            pass

        d = Dummy()
        d.context = {}
        m = messages_from_response(d)
        self.assertEqual(0, len(m))

        # object with context without messages key
        d.context = {'foo':'bar'}
        m = messages_from_response(d)
        self.assertEqual(0, len(m))

# ============================================================================

# dummy objects and settings for testing the runner
class GotHere(Exception):
    pass

def fake_loader():
    raise GotHere()

wrunner_settings = {
    'CREATE_TEMP_MEDIA_ROOT':True,
    'TEST_DATA':'awl.tests.test_waelsteng.fake_loader',
}

class WRunnerTest(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tempdir = tempfile.mkdtemp()
        cls.media_dir = os.path.abspath(os.path.join(cls.tempdir, 'media'))

    def assert_test_strings(self, expected, tests):
        names = [str(test) for test in tests]
        self.assertEqual(set(expected), set(names))

    @override_settings(WRUNNER=wrunner_settings)
    def test_runner(self):
        # this is going to get ugly... we're using the runner right now, so to
        # test the nooks and crannies we have create another one; django, as
        # of 1.11 doesn't like you to do this, so we'll mock out super() to
        # stop the parent from getting invoked

        name = '%s.super' % six.moves.builtins.__name__
        fake_runner = WRunner()
        with mock.patch(name):
            fake_runner.setup_test_environment()

            # check test data loader
            with self.assertRaises(GotHere):
                fake_runner.setup_databases()

        # check a temp media root was created
        self.assertTrue(os.path.exists(fake_runner.test_media_root))
        self.assertTrue(os.path.isdir(fake_runner.test_media_root))

        # inside of WRunner's test suites, a temp media root should have been
        # created, to test this we'll need a test suite (yes, inside of our
        # TestCase which is inside of a suite already)... told you, ugly was
        # coming

        class InteriorRunnerTestCase(TestCase):
            def runTest(self):
                from django.conf import settings
                self.assertEqual(fake_runner.test_media_root, 
                    settings.MEDIA_ROOT)

        suite = unittest.TestSuite()
        suite.addTest(InteriorRunnerTestCase())
        fake_runner.run_suite(suite)

        # -- check media root cleanup
        with mock.patch(name):
            fake_runner.teardown_databases(old_config=[])
            self.assertFalse(os.path.exists(fake_runner.test_media_root))

        # the ugliness continues! to test the building test suites we need a
        # real, not faked out, runner, so now we'll create one that isn't
        # mocked out

        real_runner = WRunner()

        # -- test various labels work
        expected = [
            'test_same_order (awl.tests.test_ranked.GroupedTests)',
            'test_same_order (awl.tests.test_ranked.AloneTests)',
            'test_too_large (awl.tests.test_ranked.GroupedTests)',
        ]
        suite = real_runner.build_suite([
            'awl.tests.test_ranked.GroupedTests.test_too_large',
            '=_same_order'])
        self.assert_test_strings(expected, suite)

        # shortcuts only
        expected = [
            'test_same_order (awl.tests.test_ranked.GroupedTests)',
            'test_same_order (awl.tests.test_ranked.AloneTests)',
        ]
        suite = real_runner.build_suite(['=_same_order'])
        self.assert_test_strings(expected, suite)

        # test no labels at all
        suite = real_runner.build_suite([])
        self.assertTrue(list(suite))

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)



# awl.logtools.py

from logthing.utils import default_logging_dict

# =============================================================================

def django_logging_dict(log_dir, handlers=['file'], filename='debug.log'):
    """Extends :func:`logthing.utils.default_logging_dict` with django
    specific values.
    """
    d = default_logging_dict(log_dir, handlers, filename)
    d['handlers'].update({
        'mail_admins':{
            'level':'ERROR',
            'class':'django.utils.log.AdminEmailHandler',
        }
    })
    d['loggers'].update({
        'django.db.backends': { # stop SQL debug from going to main logger
            'handlers': ['file', 'mail_admins'],
            'level': 'ERROR',
            'propagate': False,
        },
        'django.request': {
            'handlers': ['file', 'mail_admins'],
            'level': 'ERROR',
            'propagate': False,
        },
    })
    return d



from django.test import TestCase

from awl.logtools import django_logging_dict

# ============================================================================

class LogToolsTests(TestCase):
    def test_django_logging_dict(self):
        # not much to test here besides that it doesn't blow up
        django_logging_dict('foo')



# awl.rankedmodel.admin.py

from django.contrib.contenttypes.models import ContentType
from django.urls import reverse

# =============================================================================
# RankedModel Helper Methods
# =============================================================================

def admin_link_move_up(obj, link_text='up'):
    """Returns a link to a view that moves the passed in object up in rank.

    :param obj:
        Object to move
    :param link_text:
        Text to display in the link.  Defaults to "up"
    :returns:
        HTML link code to view for moving the object
    """
    if obj.rank == 1:
        return ''

    content_type = ContentType.objects.get_for_model(obj)
    link = reverse('awl-rankedmodel-move', args=(content_type.id, obj.id, 
        obj.rank - 1))

    return '<a href="%s">%s</a>' % (link, link_text)


def admin_link_move_down(obj, link_text='down'):
    """Returns a link to a view that moves the passed in object down in rank.

    :param obj:
        Object to move
    :param link_text:
        Text to display in the link.  Defaults to "down"
    :returns:
        HTML link code to view for moving the object
    """
    if obj.rank == obj.grouped_filter().count():
        return ''

    content_type = ContentType.objects.get_for_model(obj)
    link = reverse('awl-rankedmodel-move', args=(content_type.id, obj.id, 
        obj.rank + 1))

    return '<a href="%s">%s</a>' % (link, link_text)
